==> 1.0.0-dev.tar.gz: lib/core/flutter_type_class.dart <==
part of '../flutterx_reflection.dart';

/// Class type (e.g. 'Class String')
///
/// [token] @see [FlutterType.token]
/// [library] @see [DefinedType.library]
/// [exports] @see [DefinedType.exports]
/// [name] @see [FlutterType.name]
/// [referenceId] @see [DefinedType.referenceId]
/// [isAbstract] true if this class is abstract
/// [typeParameters] @see [RawType.typeParameters]
/// [superType] of this class or [null] if this type is [Object]
/// [interfaces] set of classes implemented by this class
/// [constructors] of this class can be used to instantiate it
/// [toString] @see [FlutterType.toString]
/// [isAssignableFrom] if other is class and is the same class or any of it's superType / interfaces are assignable to this. @see [FlutterType.isAssignableFrom]
/// [parametrize] @see [RawType.parametrize]
/// [constructor] simply return constructor by name (can be empty for unnamed constructors)
/// [newInstance] calls [FlutterConstructor.newInstance] passing this type
@immutable
abstract class FlutterClass implements FlutterType, DefinedType, RawType {
  @override
  TypeToken get token;

  @override
  String get library;

  @override
  Set<String> get exports;

  @override
  String get name;

  @override
  String get referenceId;

  bool get isAbstract;

  @override
  Set<FlutterTypeParameter> get typeParameters;

  FlutterClass? get superType;

  Set<FlutterClass> get interfaces;

  Set<FlutterConstructor> get constructors;

  @override
  String toString();

  @override
  bool isAssignableFrom(FlutterType other);

  @override
  ParametrizedFlutterClass parametrize(List<FlutterType> typeArgs);

  FlutterConstructor constructor([String name = '']);

  dynamic newInstance({required FlutterConstructor constructor, required InstanceProvider provider});
}

/// Raw implementation of [FlutterClass]
@immutable
class RawFlutterClass implements FlutterClass {
  @override
  final TypeToken token;

  @override
  final String library;

  @override
  final Set<String> exports;

  @override
  final String name;

  @override
  String get referenceId => DefinedType.generateReferenceId(this);

  @override
  final bool isAbstract;

  @override
  Set<FlutterTypeParameter> get typeParameters => _typeParameters?.call() ?? const <FlutterTypeParameter>{};

  @override
  FlutterClass? get superType => _superType?.call();

  @override
  Set<FlutterClass> get interfaces => _interfaces?.call() ?? const <FlutterClass>{};

  @override
  Set<FlutterConstructor> get constructors => _constructors?.call() ?? const <FlutterConstructor>{};

  final ValueGetter<Set<FlutterTypeParameter>>? _typeParameters;
  final ValueGetter<FlutterClass>? _superType;
  final ValueGetter<Set<FlutterClass>>? _interfaces;
  final ValueGetter<Set<FlutterConstructor>>? _constructors;

  const RawFlutterClass({
    this.token = const TypeToken<dynamic>(),
    required this.library,
    this.exports = const {},
    required this.name,
    this.isAbstract = false,
    ValueGetter<Set<FlutterTypeParameter>>? typeParameters,
    ValueGetter<FlutterClass>? superType,
    ValueGetter<Set<FlutterClass>>? interfaces,
    ValueGetter<Set<FlutterConstructor>>? constructors,
  })  : _typeParameters = typeParameters,
        _superType = superType,
        _interfaces = interfaces,
        _constructors = constructors;

  @override
  String toString() => _describeFlutterType(this);

  @override
  bool isAssignableFrom(FlutterType other) {
    if (other is FlutterTypeAlias) other = other.aliasedType; // handle type alias
    if (other is! FlutterClass) return false; // different type
    if (other.library == library && other.name == name) return true; // is the same class
    if (other.superType?.let(isAssignableFrom) == true) return true; // superType is assignable
    if (other.interfaces.any(isAssignableFrom)) return true; // interfaces are assignable
    return false;
  }

  @override
  ParametrizedFlutterClass parametrize(List<FlutterType> typeArgs) => ParametrizedFlutterClass(this, typeArgs);

  @override
  FlutterConstructor constructor([String name = '']) => constructors.require((item) => item.name == name,
      message: () => 'constructor \'$name\' not found in class $this');

  @override
  dynamic newInstance({required FlutterConstructor constructor, required InstanceProvider provider}) =>
      constructor.newInstance(provider: provider);
}

/// Parametrized implementation of [FlutterClass]
///
/// [base] @see [ParametrizedType.base]
/// [token] already parametrized basing on typeArgs. @see [FlutterClass.token]
/// [typeArgs] @see [ParametrizedType.typeArgs]
/// [isAssignableFrom] also check if typeArgs are assignable. @see [FlutterClass.isAssignableFrom]
/// [newInstance] calls [FlutterConstructor.newInstance] passing this type and typeArgs
class ParametrizedFlutterClass implements FlutterClass, ParametrizedType {
  @override
  final RawFlutterClass base;

  @override
  TypeToken get token => base.token.parametrize(typeArgs.tokens);

  @override
  String get library => base.library;

  @override
  Set<String> get exports => base.exports;

  @override
  String get name => base.name;

  @override
  String get referenceId => base.referenceId;

  @override
  bool get isAbstract => base.isAbstract;

  @override
  Set<FlutterTypeParameter> get typeParameters => base.typeParameters;

  @override
  final List<FlutterType> typeArgs;

  @override
  FlutterClass? get superType => base.superType;

  @override
  Set<FlutterClass> get interfaces => base.interfaces;

  @override
  Set<FlutterConstructor> get constructors => base.constructors;

  const ParametrizedFlutterClass(this.base, this.typeArgs);

  @override
  String toString() => _describeFlutterType(this);

  @override
  bool isAssignableFrom(FlutterType other) {
    if (other is FlutterTypeAlias) other = other.aliasedType; // handle type alias
    if (other is! ParametrizedFlutterClass) return false; // other has no type args (e.g. List<String> = List)
    if (!base.isAssignableFrom(other)) return false; // base type not assignable (e.g. List<String> = Iterable<String>)
    if (!typeArgs.areAssignableFrom(other.typeArgs)) return false; // type args not assignable
    return true;
  }

  @override
  ParametrizedFlutterClass parametrize(List<FlutterType> typeArgs) => ParametrizedFlutterClass(base, typeArgs);

  @override
  FlutterConstructor constructor([String name = '']) => base.constructor(name);

  @override
  dynamic newInstance({required FlutterConstructor constructor, required InstanceProvider provider}) =>
      constructor.newInstance(provider: provider, typeArgs: typeArgs);
}

==> 1.0.1-dev.tar.gz: lib/core/flutter_type_class.dart <==
part of '../flutterx_reflection.dart';

/// Class type (e.g. 'Class String')
///
/// [token] @see [FlutterType.token]
/// [library] @see [DefinedType.library]
/// [exports] @see [DefinedType.exports]
/// [name] @see [FlutterType.name]
/// [referenceId] @see [DefinedType.referenceId]
/// [isAbstract] true if this class is abstract
/// [typeParameters] @see [RawType.typeParameters]
/// [superType] of this class or [null] if this type is [Object]
/// [interfaces] set of classes implemented by this class
/// [constructors] of this class can be used to instantiate it
/// [fields] of this class can be used to query instance fields
/// [toString] @see [FlutterType.toString]
/// [isAssignableFrom] if other is class and is the same class or any of it's superType / interfaces are assignable to this. @see [FlutterType.isAssignableFrom]
/// [parametrize] @see [RawType.parametrize]
/// [constructor] simply return constructor by name (can be empty for unnamed constructors)
/// [field] simply return constructor by name
/// [newInstance] calls [FlutterConstructor.newInstance] passing this type
@immutable
abstract class FlutterClass implements FlutterType, DefinedType, RawType {
  @override
  TypeToken get token;

  @override
  String get library;

  @override
  Set<String> get exports;

  @override
  String get name;

  @override
  String get referenceId;

  bool get isAbstract;

  @override
  Set<FlutterTypeParameter> get typeParameters;

  FlutterClass? get superType;

  Set<FlutterClass> get interfaces;

  Set<FlutterConstructor> get constructors;

  Set<FlutterField> get fields;

  @override
  String toString();

  @override
  bool isAssignableFrom(FlutterType other);

  @override
  ParametrizedFlutterClass parametrize(List<FlutterType> typeArgs);

  FlutterConstructor constructor([String name = '']);

  FlutterField field(String name);

  dynamic newInstance({required FlutterConstructor constructor, required InstanceProvider provider});
}

/// Raw implementation of [FlutterClass]
@immutable
class RawFlutterClass implements FlutterClass {
  @override
  final TypeToken token;

  @override
  final String library;

  @override
  final Set<String> exports;

  @override
  final String name;

  @override
  String get referenceId => DefinedType.generateReferenceId(this);

  @override
  final bool isAbstract;

  @override
  Set<FlutterTypeParameter> get typeParameters => _typeParameters?.call() ?? const <FlutterTypeParameter>{};

  @override
  FlutterClass? get superType => _superType?.call();

  @override
  Set<FlutterClass> get interfaces => _interfaces?.call() ?? const <FlutterClass>{};

  @override
  Set<FlutterConstructor> get constructors => _constructors?.call() ?? const <FlutterConstructor>{};

  @override
  Set<FlutterField> get fields => _fields?.call() ?? const <FlutterField>{};

  final ValueGetter<Set<FlutterTypeParameter>>? _typeParameters;
  final ValueGetter<FlutterClass>? _superType;
  final ValueGetter<Set<FlutterClass>>? _interfaces;
  final ValueGetter<Set<FlutterConstructor>>? _constructors;
  final ValueGetter<Set<FlutterField>>? _fields;

  const RawFlutterClass({
    this.token = const TypeToken<dynamic>(),
    required this.library,
    this.exports = const {},
    required this.name,
    this.isAbstract = false,
    ValueGetter<Set<FlutterTypeParameter>>? typeParameters,
    ValueGetter<FlutterClass>? superType,
    ValueGetter<Set<FlutterClass>>? interfaces,
    ValueGetter<Set<FlutterConstructor>>? constructors,
    ValueGetter<Set<FlutterField>>? fields,
  })  : _typeParameters = typeParameters,
        _superType = superType,
        _interfaces = interfaces,
        _constructors = constructors,
        _fields = fields;

  @override
  String toString() => _describeFlutterType(this);

  @override
  bool isAssignableFrom(FlutterType other) {
    if (other is FlutterTypeAlias) other = other.aliasedType; // handle type alias
    if (other is! FlutterClass) return false; // different type
    if (other.library == library && other.name == name) return true; // is the same class
    if (other.superType?.let(isAssignableFrom) == true) return true; // superType is assignable
    if (other.interfaces.any(isAssignableFrom)) return true; // interfaces are assignable
    return false;
  }

  @override
  ParametrizedFlutterClass parametrize(List<FlutterType> typeArgs) => ParametrizedFlutterClass(this, typeArgs);

  @override
  FlutterConstructor constructor([String name = '']) => constructors.require((item) => item.name == name,
      message: () => 'constructor \'$name\' not found in class $this');

  @override
  FlutterField field(String name) =>
      fields.require((item) => item.name == name, message: () => 'field \'$name\' not found in class $this');

  @override
  dynamic newInstance({required FlutterConstructor constructor, required InstanceProvider provider}) =>
      constructor.newInstance(provider: provider);
}

/// Parametrized implementation of [FlutterClass]
///
/// [base] @see [ParametrizedType.base]
/// [token] already parametrized basing on typeArgs. @see [FlutterClass.token]
/// [typeArgs] @see [ParametrizedType.typeArgs]
/// [isAssignableFrom] also check if typeArgs are assignable. @see [FlutterClass.isAssignableFrom]
/// [newInstance] calls [FlutterConstructor.newInstance] passing this type and typeArgs
class ParametrizedFlutterClass implements FlutterClass, ParametrizedType {
  @override
  final RawFlutterClass base;

  @override
  TypeToken get token => base.token.parametrize(typeArgs.tokens);

  @override
  String get library => base.library;

  @override
  Set<String> get exports => base.exports;

  @override
  String get name => base.name;

  @override
  String get referenceId => base.referenceId;

  @override
  bool get isAbstract => base.isAbstract;

  @override
  Set<FlutterTypeParameter> get typeParameters => base.typeParameters;

  @override
  final List<FlutterType> typeArgs;

  @override
  FlutterClass? get superType => base.superType;

  @override
  Set<FlutterClass> get interfaces => base.interfaces;

  @override
  Set<FlutterConstructor> get constructors => base.constructors;

  @override
  Set<FlutterField> get fields => base.fields;

  const ParametrizedFlutterClass(this.base, this.typeArgs);

  @override
  String toString() => _describeFlutterType(this);

  @override
  bool isAssignableFrom(FlutterType other) {
    if (other is FlutterTypeAlias) other = other.aliasedType; // handle type alias
    if (other is! ParametrizedFlutterClass) return false; // other has no type args (e.g. List<String> = List)
    if (!base.isAssignableFrom(other)) return false; // base type not assignable (e.g. List<String> = Iterable<String>)
    if (!typeArgs.areAssignableFrom(other.typeArgs)) return false; // type args not assignable
    return true;
  }

  @override
  ParametrizedFlutterClass parametrize(List<FlutterType> typeArgs) => ParametrizedFlutterClass(base, typeArgs);

  @override
  FlutterConstructor constructor([String name = '']) => base.constructor(name);

  @override
  FlutterField field(String name) => base.field(name);

  @override
  dynamic newInstance({required FlutterConstructor constructor, required InstanceProvider provider}) =>
      constructor.newInstance(provider: provider, typeArgs: typeArgs);
}
